//		Copyright © 1998–1999  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: PersistentData.cpp,v $
//		Revision 1.5  1999/04/18 13:17:31  jjk
//		Get rid of virtual functions (Intel support).  Arrgh!
//
//		Revision 1.4  1999/04/18 10:12:00  jjk
//		GCC support.
//
//		Revision 1.3  1999/01/09 15:33:39  jjk
//		Explicitly include the prefix header file (R4 x86 support).
//
//		Revision 1.2  1998/06/12 09:53:41  jjk
//		Encapsulate debugging messages.
//
//		Revision 1.1  1998/06/06 15:14:30  jjk
//		Initial revision
//

//		Abstract base class for storage of persistent driver data.
//		Based on sample code provided by Trey Boudreau.

#include "Prefix.h"

#include <KernelExport.h>
#include <OS.h>

#include "PersistentData.h"

void *
PersistentData::operator new(size_t size, const char *pName, int version)
{
	// First try to find an area that we've already built.
	PersistentData *pData = locatePersistentData(pName);
	if (pData) {
		// We're done if the revisions match.
		if (pData->mVersion == version) {
			// The object must not be re-initialized.
			pData->mNeedsInitialization = false;
			return pData;
		}

		// Uh-oh, there's old data lying around.  Throw it away.
		releasePersistentData(pData);
	}

	// Create a new persistent object.
	return createPersistentData(size, pName, version);
}

void
PersistentData::operator delete(void *pData)
{
	releasePersistentData((PersistentData *)pData);
}

PersistentData *
PersistentData::locatePersistentData(const char *pName)
{
	// Look for an area that we've already built.
	const area_id theArea = find_area(pName);
	area_info info;
	if (theArea >= 0 && get_area_info(theArea, &info) == B_OK) {
		PersistentData *const pData = (PersistentData *)info.address;
		if (pData->mAreaID != theArea) {
			ddprintf(("ST - Persistent data area ID mismatch!?\n"));
		}
		return pData;
	}
	
	// Not found.
	return 0;
}

PersistentData *
PersistentData::createPersistentData(
	size_t size, const char *pName, int version)
{
	// Round size to next multiple of B_PAGE_SIZE.
	size = ((size + B_PAGE_SIZE - 1) / B_PAGE_SIZE) * B_PAGE_SIZE;
	void *pArea = 0;
	const area_id theArea = create_area(pName,
													&pArea, B_ANY_KERNEL_ADDRESS, size,
													B_FULL_LOCK, B_READ_AREA|B_WRITE_AREA);
	if (theArea >= 0) {
		PersistentData *pData = (PersistentData *)pArea;
		
		// The new object must be initialized.
		pData->mAreaID = theArea;
		pData->mNeedsInitialization = true;
		pData->mVersion = version;
		pData->mCleanup = 0;
		
		return pData;
	}
	
	// Couldn't create the area.
	return 0;
}

void
PersistentData::releasePersistentData(PersistentData *pData)
{
	if (pData->mCleanup) {
		// Allow the object to clean itself up.
		(pData->mCleanup)(pData, pData->mVersion);
	}

	// Free allocated memory.
	delete_area(pData->mAreaID);
}

static char RCSid[] = "$Id: PersistentData.cpp,v 1.5 1999/04/18 13:17:31 jjk Exp $";
