//		Copyright © 1998–1999  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: StDevices.cpp,v $
//		Revision 1.8  1999/04/18 13:17:31  jjk
//		Get rid of virtual functions (Intel support).  Arrgh!
//
//		Revision 1.7  1999/04/18 10:15:19  jjk
//		GCC support and general cleanup.
//
//		Revision 1.6  1999/01/09 15:33:39  jjk
//		Explicitly include the prefix header file (R4 x86 support).
//
//		Revision 1.5  1998/10/23 09:21:52  jjk
//		Add support for multiple partitions per physical device.
//		Adapt to new driver API (BeOS R4b1), even if we don't yet
//		support the new entry points.
//
//		Revision 1.4  1998/07/04 16:12:03  jjk
//		Move from #pragma export to the gods-rotted Microsoft method.
//
//		Revision 1.3  1998/06/22 18:46:37  jjk
//		Change device names to prepare for future partition support.
//
//		Revision 1.2  1998/06/11 22:15:35  jjk
//		Store entire device information persistently.
//
//		Revision 1.1  1998/06/06 15:17:57  jjk
//		Initial revision
//

//		Management of device names and descriptors.

#include "Prefix.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Drivers.h>
#include <KernelExport.h>
#include <OS.h>

#include "PersistentData.h"
#include "StDevices.h"
#include "StIOControl.h"
#include "StOpenClose.h"
#include "StReadWrite.h"

// Number of logical devices corresponding to a single physical device.
//
// The following logical device names will be created for each supported
// partition on each physical device found (P/T/L is replaced by the pathID,
// targetID and LUN of the device, X is replaced by a partition ID):
//
//		/dev/tape/scsi/P/T/L/X		auto rewind, compressed.
//		/dev/tape/scsi/P/T/L/Xn		AT&T style, no auto rewind, compressed.
//		/dev/tape/scsi/P/T/L/Xnb	Berkeley style, no auto rewind, compressed.
//		/dev/tape/scsi/P/T/L/Xu		auto rewind, uncompressed.
//		/dev/tape/scsi/P/T/L/Xun	AT&T style, no auto rewind, uncompressed.
//		/dev/tape/scsi/P/T/L/Xunb	Berkeley style, no auto rewind, uncompressed.
//
// The number of supported partitions is currently hardcoded.
const int kPartitionsPerPhysicalDevice = 2;
const int kLogicalDevicesPerPartition = 6;
const int kLogicalDevicesPerPhysicalDevice =
	kPartitionsPerPhysicalDevice * kLogicalDevicesPerPartition;

// Format of a device name.  Note that this string is strictly longer
// than its expansions, and so can be used to estimate memory requirements.
static const char kDeviceNameFormat[] = "tape/scsi/%d/%d/%d/%d%s%s%s";

// Table of available device names.
static const char **ppDeviceNames = 0;

// Table of device descriptors for each named device.
static LogicalDevice **ppDevices;

//----------------------------------------------------------------------------
// Placement new operator.
// We can't use the version in <new.h> because it has an exception
// specifier, which causes the stupid compiler to emit an (unreachable)
// call to __unexpected.
inline void *
operator new(size_t, void * ptr)
{
	return ptr;
}

// Allocate an array of objects and initialize it to zero.
template <class T> class allocate
{
public:
	allocate(size_t numObjects)
	:	mNumObjects(numObjects)
	{
		// empty
	}

	operator T *(void)
	{
		return (T *)calloc(mNumObjects, sizeof(T));
	}
	
private:
	const size_t mNumObjects;
};

//----------------------------------------------------------------------------
// Device information is stored in a persistent area of storage (both
// to save time when the driver is reopened and to allow user settings
// like block size to be preserved).
class PersistentDeviceInfo : public PersistentData
{
public:
	// Version number and constructor.
	enum {
		kVersion = 0
	};
	PersistentDeviceInfo(uint8 pathID, uint8 targetID, uint8 lun,
								bool removableMedium);

	// Verify that the object has been successfully initialized.
	bool IsValid(void)
	{
		return mValid;
	}

	// Return supported number of partitions for the physical device.
	int SupportedPartitions(void) const
	{
		return kPartitionsPerPhysicalDevice;
	}
		
	// Return descriptor resp. name of a given logical device.
	LogicalDevice *DeviceDescriptor(int partition, int index)
	{
		return mLogicalDevice + partition * kLogicalDevicesPerPartition + index;
	}
	
	const char *DeviceName(int partition, int index)
	{
		return mDeviceName[partition][index];
	}
	
private:
	PersistentDeviceInfo(const PersistentDeviceInfo &orig);
	PersistentDeviceInfo &operator =(const PersistentDeviceInfo &orig);

	// Clean up if version has changed.
	static void Cleanup(PersistentData *pSelf, int version);

	// Data members.
	PhysicalDevice &mPhysicalDevice;
	char mPhysDevStorage[sizeof(PhysicalDevice)];

	typedef LogicalDevice LogicalDeviceArray[kLogicalDevicesPerPhysicalDevice];
	LogicalDeviceArray &mLogicalDevice;
	char mLogDevStorage[sizeof(LogicalDeviceArray)];

	char mDeviceName[kPartitionsPerPhysicalDevice]
						 [kLogicalDevicesPerPartition]
						 [sizeof(kDeviceNameFormat)];

	bool mValid;
};

// Construct persistent device data.
PersistentDeviceInfo::PersistentDeviceInfo(
	uint8 pathID, uint8 targetID, uint8 lun, bool removableMedium)
:	mPhysicalDevice(*(PhysicalDevice *)mPhysDevStorage),
	mLogicalDevice(*(LogicalDeviceArray *)mLogDevStorage)
{
	if (NeedsInitialization()) {
		SetCleanupFunction(Cleanup);
		mValid = false;
		
		// Create the locking semaphore and assign it to the system team.
		char buf[B_OS_NAME_LENGTH];
		sprintf(buf, "SCSItape lock %d/%d/%d", pathID, targetID, lun);
		sem_id lock = create_sem(1, buf);
		if (lock >= B_OK) {
			set_sem_owner(lock, B_SYSTEM_TEAM);

			// Construct the physical device descriptor.
			new (&mPhysicalDevice)
				 PhysicalDevice(lock, pathID, targetID, lun, removableMedium);

			// Construct the logical device descriptors and device names.
			for (int i = 0; i < kPartitionsPerPhysicalDevice; ++i) {
				for (int j = 0; j < kLogicalDevicesPerPartition; ++j) {
					const bool compression = (j < 3);
					const bool autoRewind  = (j % 3) == 0;
					const bool bsdBehavior = (j % 3) == 2;
					
					new (DeviceDescriptor(i, j))
						 LogicalDevice(mPhysicalDevice, i,
						 					compression, bsdBehavior, autoRewind);
	
					sprintf(mDeviceName[i][j], kDeviceNameFormat,
							  pathID, targetID, lun,		// SCSI address
							  i,									// partition
							  compression ? "" : "u",		// flags
							  autoRewind  ? "" : "n",
							  bsdBehavior ? "b" : "");
				}
			}
			
			// Done.
			mValid = true;
		}
	}
}

void
PersistentDeviceInfo::Cleanup(PersistentData *pSelf, int /*version*/)
{
	// Free resources and invalidate the object.
	PersistentDeviceInfo &self = static_cast<PersistentDeviceInfo &>(*pSelf);
	self.mPhysicalDevice.~PhysicalDevice();
	self.mValid = false;
}

//----------------------------------------------------------------------------
// Given a map of available devices, build descriptors and assign names.
status_t
buildDeviceTables(const int nDevices, const uint8 *const pMap)
{
	status_t status = B_OK;
	
	// Create tables mapping logical device names to descriptors.
	ppDeviceNames =
		allocate<const char *>(nDevices * kLogicalDevicesPerPhysicalDevice + 1);
	if (!ppDeviceNames) {
		status = B_NO_MEMORY;
	}

	ppDevices =
		allocate<LogicalDevice *>(nDevices * kLogicalDevicesPerPhysicalDevice);
	if (!ppDevices) {
		status = B_NO_MEMORY;
	}

	// Find or create persistent data for all devices.
	int nextDevice = 0;
	const uint8 *pMapEntry = pMap;
	for (int i = 0; status == B_OK && i < nDevices; ++i) {
		while (!(*pMapEntry & kDevicePresent)) {
			++pMapEntry;
		}
		const int mapIndex = pMapEntry - pMap;
		const uint8 pathID = mapIndex / (8 * 8);
		const uint8 targetID = (mapIndex / 8) % 8;
		const uint8 lun = mapIndex % 8;
		const bool removableMedium = (*pMapEntry & kRemovableMedium) != 0;
		
		char buf[100];
		sprintf(buf, "SCSItape %d/%d/%d", pathID, targetID, lun);
		PersistentDeviceInfo *pInfo =
			new (buf, PersistentDeviceInfo::kVersion)
				 PersistentDeviceInfo(pathID, targetID, lun, removableMedium);
		
		if (!pInfo) {
			status = B_NO_MEMORY;

		} else if (!pInfo->IsValid()) {
			status = B_ERROR;

		} else {
			// Add entries to device tables.
			for (int j = 0; j < pInfo->SupportedPartitions(); ++j) {
				for (int k = 0; k < kLogicalDevicesPerPartition; ++k) {
					ppDeviceNames[nextDevice] = pInfo->DeviceName(j, k);
					ppDevices[nextDevice] = pInfo->DeviceDescriptor(j, k);
					++nextDevice;
				}
			}
		}
	}

	if (status == B_OK) {
		ddprintf(("ST - %d logical devices on %d physical drives.\n",
					 nextDevice, nDevices));

		// Store sentinel for findDevice();
		ppDeviceNames[nextDevice] = 0;

	} else {
		ddprintf(("ST - Failed to initialize %d physical devices.\n", nDevices));

		// Free any memory already allocated.
		freeDeviceTables();
	}
	
	// Done.
	return status;
}

// Free the device tables.
void
freeDeviceTables(void)
{
	if (ppDeviceNames) {
		free(ppDeviceNames);
		ppDeviceNames = 0;
	}

	if (ppDevices) {
		free(ppDevices);
		ppDevices = 0;
	}
}

// Find device descriptor for a given name.
LogicalDevice *
findDevice(const char *pName)
{
	if (ppDeviceNames) {
		for (const char **ppName = ppDeviceNames; *ppName; ++ppName) {
			if (!strcmp(*ppName, pName)) {
				return ppDevices[ppName - ppDeviceNames];
			}
		}
	}
	
	// Not found.
	return 0;
}

// Exported entry points.
PUBLIC int32 api_version = B_CUR_DRIVER_API_VERSION;

PUBLIC
const char **
publish_devices(void)
{
	return ppDeviceNames;
}

// Device hook functions.
static device_hooks kHooks = {
	stOpen, stClose, stFree, stControl, stRead, stWrite, 0, 0, 0, 0
};

PUBLIC
device_hooks *
find_device(const char *pName)
{
	if (findDevice(pName)) {
		return &kHooks;
	}
	return 0;
}

static char RCSid[] = "$Id: StDevices.cpp,v 1.8 1999/04/18 13:17:31 jjk Exp $";
