//		Copyright © 1998	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: mt.cpp,v $
//		Revision 1.2  1998/10/23 10:01:44  jjk
//		Change name of default device to conform to BeOS R4b1 conventions.
//		Print device name in usage and status messages.
//
//		Revision 1.1  1998/06/22 18:52:05  jjk
//		Initial revision
//

//		Tape control program for use with SCSItape driver.

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mtio.h"

// Table of tape operations.
static struct {
	const char *const pCommand;
	const uint32 opCode;
	const int openMode;

} kCommandTable[] = {
	{	"weof",	 MTWEOF,		O_RDWR	},
	{	"wss",	 MTWSS,		O_RDWR	},
	{	"fsr",	 MTFSR,		O_RDONLY	},
	{	"bsr",	 MTBSR,		O_RDONLY	},
	{	"fsf",	 MTFSF,		O_RDONLY	},
	{	"bsf",	 MTBSF,		O_RDONLY	},
	{	"fss",	 MTFSS,		O_RDONLY	},
	{	"bss",	 MTBSS,		O_RDONLY	},
	{	"eod",	 MTEOD,		O_RDONLY	},
	{	"rew",	 MTREW,		O_RDONLY	},
	{	"offl",	 MTOFFL,		O_RDONLY	},
	{	"eject",	 MTOFFL,		O_RDONLY	},
	{	"reten",	 MTRETENS,	O_RDONLY	},
	{	"erase",	 MTERASE,	O_RDWR	},
	{	"setblk", MTSETBSIZ,	O_RDONLY	},
	{	0,			 0,			0			}
};

// Get default tape device.
static const char *
defaultDevice(void)
{
	const char *pDevice = getenv("TAPE");
	if (!pDevice) {
		pDevice = "/dev/tape/scsi/0/2/0/0";
	}
	return pDevice;
}

// Report usage.
static void
usage(const char *pName)
{
	fprintf(stderr,
			"Usage: %s [(-f|-t) tape] command [count]\n"
			"       where <command> is one of\n"
			"          status      - report status\n"
			"          weof        - write <count> filemarks\n"
			"          wss         - write <count> setmarks\n"
			"          fsr         - space forward <count> records\n"
			"          bsr         - space backward <count> records\n"
			"          fsf         - space forward <count> filemarks\n"
			"          bsf         - space backward <count> filemarks\n"
			"          fss         - space forward <count> setmarks\n"
			"          bss         - space backward <count> setmarks\n"
			"          eod         - space to end of data\n"
			"          rew[ind]    - rewind the tape\n"
			"          offl[ine]   - rewind tape and put drive offline\n"
			"          eject       - same as 'offline'\n"
			"          reten[sion] - retension the tape\n"
			"          erase       - erase the tape\n"
			"          setblk      - set block size to <count>\n"
			"                        (0 = variable sized blocks)\n"
			"       Default <tape> is %s.\n",
			pName, defaultDevice());
}

// Perform ioctl() on a named (device) file.
static int
doIOControl(const char *pDevice, int openMode, int op, void *pArg)
{
	int fd = open(pDevice, openMode);
	if (fd >= 0) {
		int status = ioctl(fd, op, pArg);
		close(fd);
		fd = status;
	}
	if (fd < 0) {
		perror(pDevice);
	}
	return fd;
}

// Report tape status.
static const char *
densityCode(uint8 density)
{
	switch (density) {

	case kDensity800bpi:
		return "half-inch 800bpi";

	case kDensity1600bpi:
		return "half-inch 1600bpi";

	case kDensity6250bpi:
		return "half-inch 6250bpi";

	case kDensityDDS1:
		return "DDS-1";

	case kDensityDDS2:
		return "DDS-2";

	case kDensityDDS3:
		return "DDS-3";

	case kDensityDDS4:
		return "DDS-4";

	case kDensityQIC11:
		return "QIC-11";

	case kDensityQIC24:
		return "QIC-24";

	case kDensityQIC120:
		return "QIC-120";

	case kDensityQIC150:
		return "QIC-150";

	case kDensityQIC525:
		return "QIC-525";

	case kDensityQIC1350:
		return "QIC-1350";

	case kDensityQIC2100:
		return "QIC-2100";

	case kDensityQIC2GB:
		return "QIC-2GB";

	case kDensityQIC5GB:
		return "QIC-5GB";

	case kDensity8mm8200:
		return "8mm 8200";

	case kDensity8mm8500:
		return "8mm 8500";
	}
	
	return "unknown";
}

static int
reportStatus(const char *pDevice)
{
	mtget result;
	int status = doIOControl(pDevice, O_RDONLY, MTIOCGET, &result);
	if (status >= 0) {
		printf("%s is %s-line.\n",
				 pDevice,
				 GMT_ONLINE(result.mt_gstat) ? "on" : "off");

		if (GMT_WR_PROT(result.mt_gstat)) {
			printf("Medium is write protected.\n");
		}
		if (GMT_COMPRESS(result.mt_gstat)) {
			printf("Device supports data compression.\n");
		}
		uint8 d = GMT_DENSITY(result.mt_gstat);
		printf("Current density is 0x%02x (%s).\n", d, densityCode(d));

		printf("Current block size is %lu.\n", result.mt_bsize);
	}
	return status;
}

// Perform tape operation.
static int
tapeOperation(const char *pDevice, int openMode, uint32 opCode, int32 count)
{
	mtop op = { opCode, count };
	return doIOControl(pDevice, openMode, MTIOCTOP, &op);
}

// Main program.
int
main(int argc, char **ppArgv)
{
	// Get program name and default tape device.
	const char *pProgName = "mt";
	if (argc > 0) {
		pProgName = *ppArgv++;
		--argc;
	}

	const char *pDevice = defaultDevice();

	// Interpret command line.
	if (argc > 0 && (!strcmp(*ppArgv, "-f") || !strcmp(*ppArgv, "-t"))) {
		++ppArgv;
		if (--argc > 0) {
			pDevice = *ppArgv++;
			--argc;
		} else {
			usage(pProgName);
			return -1;
		}
	}
	
	if (argc == 0) {
		usage(pProgName);
		return -1;
	}
	const char *pCommand = *ppArgv++;
	--argc;
	
	int32 count = 1;
	if (argc > 1) {
		usage(pProgName);
		return -1;
	} else if (argc > 0) {
		count = atol(*ppArgv);
	}
	
	// Check & perform command.
	if (!strcmp(pCommand, "status")) {
		return reportStatus(pDevice);
	}

	int i;
	const char *pCmp;
	for (i = 0; (pCmp = kCommandTable[i].pCommand) != 0; ++i) {
		if (!strncmp(pCommand, pCmp, strlen(pCmp))) {
			break;
		}
	}
	if (pCmp) {
		return tapeOperation(pDevice,
									kCommandTable[i].openMode,
									kCommandTable[i].opCode,
									count);
	}
	
	usage(pProgName);
	return -1;
}

static char RCSid[] = "$Id: mt.cpp,v 1.2 1998/10/23 10:01:44 jjk Exp $";
